'''
.. epigraph:: Functions for working with :class:`pandas.DataFrame`\\ s
'''

import numpy as np
import pandas as pd

from .cleaning_functions import clean_null, fingerprint


def find_column_labels(
    df, label_fingerprints, label_match_thresh=3, special_characters=''
):
    """
    Finds a row of column labels within a :class:`pandas.DataFrame` based on
    a collection of expected ``label_fingerprints``.

    ``df.columns`` is updated to contain the found column labels and all rows
    before and including the column label row are removed. If the initial
    column labels of ``df`` fit the match criteria, ``df`` will be unchanged.

    .. note::
       All labels must be in the same row. This function will not attempt to
       correct translational errors.

    Usage:
      >>> import pandas as pd
      >>> from etl_toolbox.dataframe_functions import find_column_labels
      >>> df = pd.DataFrame(
      ...     [
      ...         ["created by:", "etl-toolbox",             "",         "-"],
      ...         [ "2020-06-07",         "---",       "3 rows", "4 columns"],
      ...         [           "",            "",             "",         "-"],
      ...         [      "email",        "date",        "phone",        "id"],
      ...         ["aaa@aaa.com",     "04mar14", "999-333-4444",       "AAA"],
      ...         ["baa@baa.com",     "05aug13", "111-222-3333",       "BAA"],
      ...         ["caa@caa.com",     "01jun15", "777-777-7777",       "CAA"]
      ...     ]
      ... )
      >>> print(df)
                   0            1             2          3
      0  created by:  etl-toolbox                        -
      1   2020-06-07          ---        3 rows  4 columns
      2                                                  -
      3        email         date         phone         id
      4  aaa@aaa.com      04mar14  999-333-4444        AAA
      5  baa@baa.com      05aug13  111-222-3333        BAA
      6  caa@caa.com      01jun15  777-777-7777        CAA
      >>> label_fingerprints = {'email', 'date', 'phone'}
      >>> find_column_labels(df, label_fingerprints)
      >>> print(df)
               email     date         phone   id
      0  aaa@aaa.com  04mar14  999-333-4444  AAA
      1  baa@baa.com  05aug13  111-222-3333  BAA
      2  caa@caa.com  01jun15  777-777-7777  CAA

    :param df:
        A :class:`pandas.DataFrame` containing column labels as a row, possibly
        preceded by misc. non-data rows.

    :param label_fingerprints:
        Fingerprinted label names that are expected in the column labels row.
        It can contain many variations on the expected label names, which
        makes this function useful for cleaning multiple files with varied
        column labels.

        Example:
          >>> label_fingerprints = {'email', 'emailaddr', 'phone', 'phonenum'}

        .. note::
           If a `dict` is provided, the keys will be used as the fingerprints
           and the values will be ignored.

        A ``fingerprint_map`` dictionary created for use in the
        :mod:`~etl_toolbox.mapping_functions` module can be passed as
        ``label_fingerprints``.

    :type label_fingerprints: set, list, or dict

    :param label_match_thresh:
        The number of fingerprints that must be found in
        ``label_fingerprints`` for a row to be identified as the label row.
        Default is ``3``.

        This exists to prevent false matches caused by non-data header rows
        containing values that are also expected column labels.

        For example, in the following :class:`~pandas.DataFrame`, row 1 will be
        misidentified as the label row if the ``label_match_thresh`` is set
        to ``1``:

          >>> import pandas as pd
          >>> from etl_toolbox.dataframe_functions import find_column_labels
          >>> df = pd.DataFrame(
          ...     [
          ...         [      "name:", "data report",             "",         "-"],
          ...         [      "date:",  "2001-05-03",       "3 rows", "4 columns"],
          ...         [           "",            "",             "",         "-"],
          ...         [      "email",        "date",        "phone",        "id"],
          ...         ["aaa@aaa.com",     "04mar14", "999-333-4444",       "AAA"],
          ...         ["baa@baa.com",     "05aug13", "111-222-3333",       "BAA"],
          ...         ["caa@caa.com",     "01jun15", "777-777-7777",       "CAA"]
          ...     ]
          ... )
          >>> print(df)
                       0            1             2          3
          0        name:  data report                        -
          1        date:   2001-05-03        3 rows  4 columns
          2                                                  -
          3        email         date         phone         id
          4  aaa@aaa.com      04mar14  999-333-4444        AAA
          5  baa@baa.com      05aug13  111-222-3333        BAA
          6  caa@caa.com      01jun15  777-777-7777        CAA
          >>> label_fingerprints = {'email', 'date', 'phone'}
          >>> find_column_labels(df, label_fingerprints, label_match_thresh=1)
          >>> print(df)
                   date: 2001-05-03        3 rows 4 columns
          0                                               -
          1        email       date         phone        id
          2  aaa@aaa.com    04mar14  999-333-4444       AAA
          3  baa@baa.com    05aug13  111-222-3333       BAA
          4  caa@caa.com    01jun15  777-777-7777       CAA

        ``3`` is a good value for most datasets with a known set of
        ``label_fingerprints``. It can be set lower if the incoming data has
        few columns and/or highly varied label names.

        If this argument is set to ``0``, the function will raise a
        :exc:`ValueError`.

    :type label_match_thresh: int, optional

    :param special_characters:
        (optional) A string of special characters to preserve while creating
        the fingerprints for lookup in ``label_fingerprints``. See
        :func:`cleaning_functions.fingerprint()
        <etl_toolbox.cleaning_functions.fingerprint>` for details.

        .. note::
           Any special characters that appear in the elements of
           ``label_fingerprints`` should be included here.

    :type special_characters: string, optional

    :raises IndexError:
        Raised if a label row can not be identified in the given
        :class:`~pandas.DataFrame`.

    :raises ValueError:
        Raised if the ``label_match_thresh`` is set to `0`.

    :return:
        Returns ``None``. The ``df`` argument is mutated.
    """
    if label_match_thresh == 0:
        raise ValueError("label_match_thresh can not be 0.")

    # First, check if the initial labels are already correct. If they are,
    # exit without changing df.

    label_count = 0

    for cell in df.columns:
        cell_fingerprint = fingerprint(
            cell, special_characters=special_characters
        )

        if cell_fingerprint in label_fingerprints:
            label_count += 1

    
    if label_count >= label_match_thresh:
        return None

    # Check whether the index is something other than the default RangeIndex,
    # and temporarily change it to a RangeIndex if it is.
    #
    # This is necessary because pandas doesn't provide a way to drop rows in
    # place using integer locations, however using the named index will produce
    # unexpected results if the DataFrame has duplicate index values.
    initial_index = df.index
    initial_index_is_default = index_is_default(df)

    if not initial_index_is_default:
        df.reset_index(drop=True, inplace=True)

    # Iterate over rows to find the label index
    label_index = None

    for i, row in df.iterrows():
        # Count the number of fingerprinted cell values found in
        # label_fingerprints for this row
        label_count = 0

        for cell in row.iteritems():
            cell_fingerprint = fingerprint(
                cell[1], special_characters=special_characters
            )

            if cell_fingerprint in label_fingerprints:
                label_count += 1

        # When label row is found, record the index and break
        if label_count >= label_match_thresh:
            label_index = i
            break

    if label_index is None:
        raise IndexError(
            'Label row could not be identified. Make sure '
            'label_fingerprints contains the expected label names.'
        )

    # Set DataFrame column labels
    df.rename(columns=df.loc[label_index], inplace=True)

    # Remove rows up to and including the label index
    df.drop(df.loc[:label_index].index, inplace=True)

    # If the initial index was a default RangeIndex, reset it so it is
    # numbered from 0.
    # Otherwise, restore the intial index, sliced to line up with the
    # modified DataFrame.
    if initial_index_is_default:
        df.reset_index(drop=True, inplace=True)
    else:
        df.index = initial_index[label_index + 1:]


def merge_columns_by_label(df, deduplicate_values=False):
    """
    Merges columns of a :class:`pandas.DataFrame` that have identical labels

    For duplicate column labels in ``df``, the first instance of each label
    will be turned into a column of lists containing the values from all of
    the instances. The other instances will then be dropped.

    ``None`` and ``np.nan`` values will not be included in the merged column.

    .. note::
       This function does not fingerprint the labels for comparison. Column
       labels should be cleaned and mapped before using this tool.

    Usage:
      >>> import pandas as pd
      >>> from etl_toolbox.dataframe_functions import merge_columns_by_label
      >>> df = pd.DataFrame(
      ...     [
      ...         ["AAA", "aaa@aaa.com", "111-111-1111", "111@aaa.com"],
      ...         ["BAA", "baa@baa.com", "222-222-2222", "222@baa.com"],
      ...         ["CAA", "caa@caa.com", "333-333-3333", None],
      ...         ["DAA", "daa@daa.com", "444-444-4444", "444@daa.com"]
      ...     ],
      ...     columns=["id", "email", "phone", "email"]
      ... )
      >>> print(df)
          id        email         phone        email
      0  AAA  aaa@aaa.com  111-111-1111  111@aaa.com
      1  BAA  baa@baa.com  222-222-2222  222@baa.com
      2  CAA  caa@caa.com  333-333-3333         None
      3  DAA  daa@daa.com  444-444-4444  444@daa.com
      >>> merge_columns_by_label(df)
      >>> print(df)
          id                       email         phone
      0  AAA  [aaa@aaa.com, 111@aaa.com]  111-111-1111
      1  BAA  [baa@baa.com, 222@baa.com]  222-222-2222
      2  CAA               [caa@caa.com]  333-333-3333
      3  DAA  [daa@daa.com, 444@daa.com]  444-444-4444

    :param df:
        A :class:`pandas.DataFrame`.

    :param deduplicate_values:
        If ``True``, the values of the combined columns will be deduplicated
        and stored in the modified :class:`~pandas.DataFrame` as a `set`
        instead of a `list`.

    :type deduplicate_values: boolean, optional

    :return:
        Returns ``None``. The ``df`` argument is mutated.
    """

    # Get the set of duplicate column labels
    duplicate_labels = set(df.columns[df.columns.duplicated()])

    for label in duplicate_labels:
        # Get integer location of first instance of label
        first_iloc = next(i for i, x in enumerate(df.columns) if x == label)

        # Make a new column that combines the values for all instances of label
        temp_column = []

        for cell in df[label].values:
            cell = [x for x in cell if not pd.isnull(x)]  # Remove None and nan

            if deduplicate_values:
                temp_column.append(set(cell))
            else:
                temp_column.append(cell)

        df.insert(first_iloc, 'temp', temp_column)

        # Drop the duplicate columns
        temp_labels = df.columns.tolist()
        temp_labels[first_iloc] = 'temp'

        if label is None:
            # drop() can't take None as a column name, so this is a workaround.
            # As long df doesn't contain this exact label name, it will work
            # as expected.
            placeholder_label = '**be15492e-d01e-4ab9-9f2c-116525a26897**-label'
            df.columns = [
                placeholder_label if x is None else x for x in df.columns
            ]
            df.drop(columns=placeholder_label, inplace=True)
        else:
            df.drop(columns=label, inplace=True)

        # Name the new column back to the original label
        temp_labels = df.columns.tolist()
        temp_labels[first_iloc] = label
        df.columns = temp_labels


def index_is_default(df):
    """
    Returns ``True`` if the provided :class:`~pandas.DataFrame` has the
    default :class:`pandas.RangeIndex`.

    Else returns ``False``.
    """

    return df.index.equals(pd.RangeIndex(df.shape[0])) and df.index.name is None


def dataframe_clean_null(
    df,
    empty_row_thresh=1,
    empty_column_thresh=1,
    falsey_is_null=False,
    special_characters='',
):
    """
    Cleans null values of a :class:`pandas.DataFrame` and removes empty
    rows/columns.

    .. warning::
       This function is computationally intensive and might be slow on large
       :class:`~pandas.DataFrame`\\ s.

    Usage:
      >>> import pandas as pd
      >>> from etl_toolbox.dataframe_functions import dataframe_clean_null
      >>> df = pd.DataFrame(
      ...     [
      ...         ["AAA",          None, "111-111-1111", "empty"],
      ...         ["BAA", "baa@baa.com",            "-",     "-"],
      ...         ["CAA", "caa@caa.com", "notavailable",   "..."],
      ...         ["DAA",     "blocked", "444-444-4444",  "null"]
      ...     ],
      ...     columns=["id", "email", "phone", "col4"]
      ... )
      >>> print(df)
          id        email         phone   col4
      0  AAA         None  111-111-1111  empty
      1  BAA  baa@baa.com             -      -
      2  CAA  caa@caa.com  notavailable    ...
      3  DAA      blocked  444-444-4444   null
      >>> dataframe_clean_null(df)
      >>> print(df)
          id        email         phone
      0  AAA          NaN  111-111-1111
      1  BAA  baa@baa.com           NaN
      2  CAA  caa@caa.com           NaN
      3  DAA          NaN  444-444-4444

    :param df:
        A :class:`pandas.DataFrame`.

    :param empty_row_thresh:
        The number of non-null values required for a row to be considered
        populated/non-empty. Default is ``1``.

        If set to ``0``, no rows will be removed.

    :type empty_row_thresh: int, optional

    :param empty_column_thresh:
        The number of non-null values required for a column to be considered
        populated/non-empty. Default is ``1``.

        .. note::
           Rows are dropped before columns, so this threshold will be applied
           to the values that remain **after rows are removed**. If
           ``empty_column_thresh`` is greater than ``1``, the resulting ``df``
           may contain rows with fewer populated cells than
           ``empty_row_thresh``. However, it will never contain completely
           empty rows (unless ``empty_row_thresh`` is ``0``).

        If set to ``0``, no columns will be removed.

    :type empty_column_thresh: int, optional

    :param falsey_is_null:
        Controls whether falsey objects are considered *null-indicating*.
        See :func:`cleaning_functions.clean_null()
        <etl_toolbox.cleaning_functions.clean_null>` for details.
        Default is ``False``.

    :type falsey_is_null: boolean, optional

    :param special_characters:
        A string of special characters to preserve while creating the
        fingerprints. See :func:`cleaning_functions.fingerprint()
        <etl_toolbox.cleaning_functions.fingerprint>` for details.

    :type special_characters: string, optional

    :return:
        Returns ``None``. The ``df`` argument is mutated.
    """

    initial_index_is_default = index_is_default(df)

    # Apply clean_null() to every cell in df
    for i in range(df.shape[0]):
        for j in range(df.shape[1]):
            val = clean_null(
                df.iloc[i, j],
                falsey_is_null=falsey_is_null,
                special_characters=special_characters,
            )
            if val is None:
                df.iloc[i, j] = np.nan

    # Drop rows with fewer populated cells than empty_row_thresh
    df.dropna(axis=0, thresh=empty_row_thresh, inplace=True)

    # Drop columns with fewer populated cells than empty_column_thresh
    df.dropna(axis=1, thresh=empty_column_thresh, inplace=True)

    # Make sure there are no empty rows in the final df
    # (unless empty_row_thresh is 0)
    if empty_column_thresh > 1 and empty_row_thresh != 0:
        df.dropna(axis=0, how='all', inplace=True)

    # Reset the index if it was initially a default index
    if initial_index_is_default:
        df.reset_index(drop=True, inplace=True)



'''
.. epigraph:: Functions for working with files and directories
'''

import os
import re


def get_file_list_from_dir(dir_path, recursive=False, include_regex=None):
    r"""
    Returns a list of the files in a directory

    Usage:
      >>> get_file_list_from_dir('test_data/test_dir') # doctest:+SKIP
      ['test_data/test_dir/1.csv',
       'test_data/test_dir/2.csv',
       'test_data/test_dir/3.json']

    :param recursive:
        If set to ``True``, the returned list will include files from ``dir``
        and all of its subdirectories. Default is ``False``.

        Example:
          >>> get_file_list_from_dir('test_data/test_dir',
          ...                        recursive=True)  # doctest:+SKIP
          ['test_data/test_dir/1.csv',
           'test_data/test_dir/2.csv',
           'test_data/test_dir/3.json',
           'test_data/test_dir/a/1.csv',
           'test_data/test_dir/b/3.csv',
           'test_data/test_dir/b/c/2.txt']

    :type recursive: boolean, optional

    :param include_regex:
        Only include files whose path matches this regex. Default is ``None``
        (list is unfiltered).

        Example:
          >>> get_file_list_from_dir('test_data/test_dir',
          ...                        include_regex=r'.*\.csv$') # doctest:+SKIP
          ['test_data/test_dir/1.csv',
           'test_data/test_dir/2.csv']

    :type include_regex: string, optional

    :return:
        Returns list of file paths.
    """

    file_list = []

    # Collect file paths
    for root, dirs, files in os.walk(dir_path):
        for f in files:
            file_list.append(os.path.normpath(os.path.join(root, f)))

        if not recursive:
            break

    # Apply filter
    if include_regex is not None:
        file_list = [f for f in file_list if re.match(include_regex, f)]

    return file_list



import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

import etl_toolbox

with open("README.rst", "r") as r:
    long_description = r.read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ["--doctest-modules",
                          "--cov=etl_toolbox",
                          "--doctest-glob=*.rst",
                          "--ignore=docs/conf.py"]
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


setup(
    name='etl-toolbox',
    version=etl_toolbox.__version__,
    author='Brooklyn Rose Ludlow',
    description='Useful ETL functions for Python',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://etl-toolbox.readthedocs.io',
    project_urls={
        "Source Code": "https://github.com/brookcub/etl-toolbox",
        "Bug Reports": "https://github.com/brookcub/etl-toolbox/issues",
        "Documentation": "https://etl-toolbox.readthedocs.io",
    },
    keywords='etl pandas data cleaning',
    license='Apache-2.0',
    packages=['etl_toolbox'],
    install_requires=['numpy>=1.18.0',
                      'pandas>=0.25.0'
                      ],
    python_requires='>=3.5',
    tests_require=['pytest',
                   'coverage',
                   'pytest-cov'
                   ],
    cmdclass={'test': PyTest},
    platforms='any',
    test_suite='etl_toolbox.test.test_etl_toolbox',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],
    extras_require={
        'testing': ['pytest'],
    }
)



'''
.. epigraph:: Basic transformations for cleaning individual units of data
'''

import ast
import collections.abc
import re


def fingerprint(x, special_characters=''):
    """
    Returns a lowercase, alphanumeric representation of ``x``

    Usage:
      >>> from etl_toolbox.cleaning_functions import fingerprint
      >>> fingerprint('(Aa_Bb_Cc)')
      'aabbcc'

    :param x:
        The object to be fingerprinted. Will be cast to a string using
        ``str(x)``.

    :param special_characters:
        A string of special characters to preserve while creating the
        fingerprint of ``x``. Any special characters which are individually
        meaningful for the data should be included here.

          >>> fingerprint('(Aa_Bb_Cc)', special_characters='_')
          'aa_bb_cc'

        .. note::
           An example of where this might be useful is when fingerprinting
           column labels on a data set. Some datasets use ``$`` as a label for
           price/cost or ``#`` as a label for phone number or id number. In
           that case, ``special_characters`` should be set to ``'#$'`` so that
           those values are preserved.

    :type special_characters: string, optional

    :return:
        Returns a string.
    """
    remove_regex = r'[^0-9a-z{}]'.format(re.escape(special_characters))

    return re.sub(remove_regex, '', str(x).lower())


#: A list of strings that are considered equivalent to ``None``. Used by
#: :func:`clean_null()`.
NULL_INDICATORS = [
    '',
    'blank',
    'blocked',
    'empty',
    'invalid',
    'na',
    'nan',
    'nbsp',
    'none',
    'notavailable',
    'notset',
    'np',
    'null',
    'removed',
    'unavailable',
    'unidentified',
    'unknown'
]

#: A list of strings that are considered to be falsey.
#: Used by :func:`clean_null()`.
FALSEY_INDICATORS = [
    'false',
    '0'
]


def clean_null(x, falsey_is_null=False, special_characters=''):
    """
    Returns ``None`` if ``x`` is *null-indicating*, else returns ``x``

    ``x`` is considered *null-indicating* if any of the following are true:

    - ``x is None``
    - ``len(x) == 0``
    - ``fingerprint(x) in NULL_INDICATORS`` (see: :const:`NULL_INDICATORS`)
    - ``falsey_is_null and !x``
    - ``falsey_is_null and fingerprint(x) in FALSEY_INDICATORS`` (see:
      :const:`FALSEY_INDICATORS`)
    - ``x`` is an iterable consisting of all *null-indicating* values \n
      - Ex: ``x == ['empty', None, {None}]``
    - ``x`` evaluates as a Python literal that is *null-indicating* \n
      - Ex: ``x == '{None, None}'``

    Usage:
      >>> from etl_toolbox.cleaning_functions import clean_null
      >>> clean_null('Unknown') is None
      True
      >>> clean_null(['empty', None, {None}]) is None
      True
      >>> clean_null('false') is None
      False
      >>> clean_null('false', falsey_is_null=True) is None
      True

    :param x:
        The object to be evaluated. ``x`` can be any type, though this function
        is intended for use with strings, lists, and sets.

        .. warning::
            The behavior of :func:`clean_null()` may not be intuitive for types
            other than `string`, `list`, and `set`. For example, a `dict` will
            be considered *null-indicating* if it is either empty, or all of
            its **keys** are *null-indicating*. **Dictionary values will be
            ignored.**

    :param falsey_is_null:
        Controls whether falsey objects are considered *null-indicating*.

        - If ``True``, all falsey values will be considered *null-indicating*
        - If ``False``, falsey values will not be considered *null-indicating*
          (unless they are otherwise *null-indicating* regardless of falsiness)

        Default is ``False``.

        .. note::
            A common use case for setting ``falsey_is_null`` to ``True`` would
            be when evaluating non-numeric data such as locations or phone
            numbers. In these data sets, ``0`` is not a meaningful value and is
            likely just functioning as a placeholder for null.

    :type falsey_is_null: boolean, optional

    :param special_characters:
        A string of special characters to preserve while creating the
        fingerprint of ``x``. See :func:`fingerprint()` for more details.

    :type special_characters: string, optional

    :return:
        Returns ``None`` or ``x``.
    """

    if x is None:
        return None

    # Check if x is a Sized object with length 0
    if isinstance(x, collections.abc.Sized) and len(x) == 0:
        return None

    # Check if fingerprint of x is in NULL_INDICATORS
    x_fingerprint = fingerprint(x, special_characters)
    if x_fingerprint in NULL_INDICATORS:
        return None

    # Optional falsey_is_null checks
    if falsey_is_null:
        if not x:
            return None
        if x_fingerprint in FALSEY_INDICATORS:
            return None

    # Check if x is an iterable consisting of all null-indicating values
    #     Strings are excluded to improve performance. They will
    #     never be null-indicating in this check.
    if isinstance(x, collections.abc.Iterable) and not isinstance(x, str):
        contains_non_null = False

        for item in x:
            if clean_null(item, falsey_is_null) is not None:
                contains_non_null = True
                break

        if not contains_non_null:
            return None

    # Check if x evaluates as a Python literal that is null-indicating
    #     Limit to strings that are parseable to list/set/dict/tuple
    #     or special string literals to improve performance. Any
    #     other strings will never be null-indicating in this check.
    if isinstance(x, str) and (x[:1] in ('[','{','(') or x[1:2] in ('"',"'")):
        try:
            x_eval = ast.literal_eval(x)
            if clean_null(x_eval, falsey_is_null) is None:
                return None
        except Exception:
            pass

    # Return x if is has not been found null-indicating in the above checks
    return x


def clean_whitespace(x):
    """
    Returns ``x`` with whitespace characters trimmed and condensed

    Cleaning rules:
     - all whitespace characters replaced with standard ASCII space (32)
     - consecutive whitespace condensed
     - leading/trailing whitespace removed

    Usage:
      >>> from etl_toolbox.cleaning_functions import clean_whitespace
      >>> clean_whitespace(''' 123   abc 456
      ...                               def\t\t 789\t''')
      '123 abc 456 def 789'

    :param x:
        The string to be cleaned.
        .. note::
            If ``x`` is not a string, it will be returned unchanged.

    :return:
        Returns the same type as ``x``.

    .. note::
        :func:`clean_whitespace()` does not remove unicode formatting
        characters without the **White_Space** character property:

        .. hlist::
           :columns: 3

           - **U+180E**
           - **U+200B**
           - **U+200C**
           - **U+200D**
           - **U+2060**
           - **U+FEFF**
    """

    if not isinstance(x, str):
        return x

    return " ".join(x.split())



'''
.. epigraph:: Functions for mapping collections of values
'''

from .cleaning_functions import fingerprint


def map_labels(
    labels, fingerprint_map, special_characters='', return_unmapped=False
):
    """
    Maps a list of ``labels`` to new values based on provided
    ``fingerprint_map`` and returns that mapped list.

    This is useful for mapping column labels from a dataframe/file to a
    standard schema, particularly when the labels are inconsistent.

    The order of ``labels`` will be preserved in the return, and if a label
    isn't found in ``fingerprint_map``, that label will be ``'-'`` in
    returned list.

    Usage:
      >>> from etl_toolbox.mapping_functions import map_labels
      >>> labels = [1, '2_A', '2b']
      >>> fingerprint_map = {'1': 'one', '2a': 'two_a', 'extrakey': 'value'}
      >>> map_labels(labels, fingerprint_map)
      ['one', 'two_a', '-']

    :param labels:
        The list of labels to map. These will be fingerprinted for their
        lookup in ``fingerprint_map``.

    :param fingerprint_map:
        A dictionary of all expected label fingerprints mapped to formatted
        outputs.

    :param special_characters:
        A string of special characters to preserve while fingerprinting the
        labels. See :func:`cleaning_functions.fingerprint()
        <etl_toolbox.cleaning_functions.fingerprint>` for details.

        .. note::
           This should include any special characters that appear in the keys
           of ``fingerprint_map``.

    :type special_characters: string, optional

    :param return_unmapped:
        If this is set to ``True``, this function will return a tuple of the
        mapped labels and a set of unmapped labels (any value from ``labels``
        whose fingerprint was not found in ``fingerprint_map``). Default is
        ``False``.

        .. note::
           This is useful for tracking unrecognized labels of incoming files in
           an automated ETL system.

    :type return_unmapped: boolean, optional

    :return:
        Returns a list or, if the ``return_unmapped`` option is ``True``,
        returns a tuple, with the first element being a list and the second
        being a set.
    """
    mapped_labels = []
    unmapped_labels = set()

    for x in labels:
        x_fingerprint = fingerprint(x, special_characters=special_characters)

        if x_fingerprint in fingerprint_map:
            x_mapped = fingerprint_map[x_fingerprint]
            mapped_labels.append(x_mapped)
        else:
            mapped_labels.append('-')
            unmapped_labels.add(x)

    if return_unmapped:
        return (mapped_labels, unmapped_labels)

    return mapped_labels


def append_count(x):
    """
    A generator function that yields ``x`` with a numbered suffix.
    """
    i = 0
    while True:
        i += 1
        yield x + '_' + str(i)


def rename_duplicate_labels(labels, rename_generator=append_count):
    """
    Maps a list of ``labels`` such that duplicates are renamed according to the
    ``rename_generator``.

    The order of ``labels`` is preserved in the return, and if a label isn't a
    duplicate, its value will be unchanged. Values will NOT be fingerprinted
    for comparison, so this function is best used after labels have been
    standardized.

    Usage:
      >>> from etl_toolbox.mapping_functions import rename_duplicate_labels
      >>> labels = ['email', 'email', 'phone', 'name', 'email', 'phone']
      >>> rename_duplicate_labels(labels)
      ['email_1', 'email_2', 'phone_1', 'name', 'email_3', 'phone_2']

    :param labels:
        The list of labels to map.

    :param rename_generator:
        A generator function that specifies how to rename duplicate columns. It
        should take a label name as a positional argument and yield the renamed
        label. The default ``rename_generator`` appends a count, separated by
        underscore.

        Example:
          >>> def rename_generator(x):
          ...     i = 0
          ...     while True:
          ...         i += 1
          ...         yield x + '_' + str(i)
          >>> r = rename_generator('label')
          >>> next(r)
          'label_1'
          >>> next(r)
          'label_2'

    :type rename_generator: generator, optional

    :return:
        Returns a list.
    """

    # Create dictionary of duplicate labels with initialized rename_generators
    seen = set()
    duplicates = {}
    for x in labels:
        if x in seen:
            duplicates[x] = rename_generator(x)
        else:
            seen.add(x)

    # Create a new list with each duplicate label renamed according to its
    # rename_generator
    mapped_labels = []
    for x in labels:
        if x in duplicates:
            x_renamed = next(duplicates[x])
            mapped_labels.append(x_renamed)
        else:
            mapped_labels.append(x)

    return mapped_labels
